/* serve.c -- Serve the data to a connected client.
 * By: Steve Conley, tanabi@bellsouth.net, sdconley@eos.ncsu.edu
 */
 
/* System includes */
#include <sys/types.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <dirent.h>

/* Project includes */
#include "nettool/nettool.h"

/* Grab the connection list */
extern	INPCON*	conhead;
extern	char*	myname;

/* Local functions gone global */
void		process(CON* con);
FILE*	outp=stdout;

/* Local functions still at home */
void filetype(char* path,CON* con); /* This determines the content-type of a file */
/* This is to process directories */
void outdir(char* path,CON* con,int httpver);

void process(CON* con){
	char*	buf;
	time_t	timeout;
	int		done,httpver,i,status,err,j,isdir;
	char		path[8192];
	char*	cwd;
	FILE*	in;
/* First, let's wait for our friend on the other side to send us something.
 * We want to fix a timeout so that we don't get stuck forever.
 *
 * I chose a 10 second timeout; that's probably a little long in the
 * `real world' where we should have the request data -imedeatly-, but
 * just in case a grader wants to telnet in...
 */
	timeout=time(0)+10; /* 10 second timeout */
	done=0;
	while((timeout > time(0))&&(!done)){
		buf=conrecv(con); /* If we can't get anything in 10 seconds, kill
		                   * the connect and go back and get ready for a
					    * new one.
					    */
		if(buf)
			done=1;
	}
	if(!done){
		disconnect(conhead,con);
		return;
	}
	status=200; /* A-Ok! */
	err=0;
	if(strncmp(buf,"GET",3)){ /* We expect GET in all caps. */
		status=501; /* Not implemented */
		err=1;
	}
/* Clear out spaces */
	if(!err){
		for(i=3; (isspace(buf[i]))&&(i < strlen(buf)); i++){}
		if(i == strlen(buf)){
			status=404; /* File not found */
			err=1;
		}
	}
	if(!err){ /* If we got a path, process it and find out if it's a dir */
		bzero(path,8192);
		j=0;
		for(; (!isspace(buf[i]))&&(i <= strlen(buf)); i++,j++)
			path[j]=buf[i];
		i++;
		if(buf[i]!='H') /* .9 or 1.*? */
			httpver=0;
		else
			httpver=1;
		cwd=getcwd(NULL,8192); /* Append the CWD to the front of the thing */
		strcat(cwd,"/");
		strcat(cwd,path);
		strcpy(path,cwd);
		free(cwd);
		fprintf(outp,"%s: Requested file %s\n",con->host,path);
	}
	free(buf);
/* If it's a dir (ie no file given) process it as such. */
	if(path[strlen(path)-1]=='/'){
		outdir(path,con,httpver);
		return;
	}
/* We now have the path that we want to see */
	if(!httpver){ /* If we're processing a HTTP .9 request */
		if(err){
			sprintf(path,
"<HTML><TITLE>Error!</TITLE><BODY>\r\n"
"<H2><CENTER>I'm sorry, but I recieved an error -- Error number %i</H2></CENTER>\r\n"
"Refer to this chart to see what it means:\r\n<HR>\r\n"
"<TABLE width=100%%>"
"<TR><TD>501</TD><TD>Not Implemented.  Your browser made a request I don't understand.</TD></TR>\r\n"
"<TR><TD>404</TD><TD>File Not Found (does the web page exist anymore?)</TD></TR>\r\n"
"</TABLE></BODY></HTML>\r\n",status,status);
			consend(con,path);
			disconnect(conhead,con);
			return;
		}
		in=fopen(path,"r");
		if(!in){
			consend(con,
"<HTML><TITLE>Error!</TITLE><BODY>\r\n"
"<H2><CENTER>I'm sorry, but I recieved an error -- Error number 404</H2></CENTER>\r\n"
"This means that the file you were searching for wasn't found on this server.\r\n"
"Please make sure that your URL is correct and that the file exists.\r\n"
"</BODY></HTML>\r\n");
			disconnect(conhead,con);
			return;
		}
	}else{ /* Otherwise, let's expect more arguements */
/* Currently, let's pretend we don't understand 'em */
		if(err){
			sprintf(path,
"HTTP/1.0 %i Error\r\n"
"Content-type: text/html\r\n"
"Connection: close\r\n"
"\r\n"
"<HTML><TITLE>Error!</TITLE><BODY>\n"
"<H2><CENTER>I'm sorry, but I recieved an error -- Error number %i</H2></CENTER>\r\n"
"Refer to this chart to see what it means:\r\n<HR>\r\n"
"<TABLE width=100%%>"
"<TR><TD>501</TD><TD>Not Implemented.  Your browser made a request I don't understand.</TD></TR>\r\n"
"<TR><TD>404</TD><TD>File Not Found (does the web page exist anymore?)</TD></TR>\r\n"
"</TABLE></BODY></HTML>\r\n",status,status);
			consend(con,path);
			disconnect(conhead,con);
			return;
		}
		in=fopen(path,"r");
		if(!in){
			consend(con,
"HTTP/1.0 404 Not Found\r\n"
"Content-type: text/html\r\n"
"Connection: close\r\n"
"\r\n"
"<HTML><TITLE>Error!</TITLE><BODY>\r\n"
"<H2><CENTER>I'm sorry, but I recieved an error -- Error number 404</H2></CENTER>\r\n"
"This means that the file you were searching for wasn't found on this server.\r\n"
"Please make sure that your URL is correct and that the file exists.\r\n");
			disconnect(conhead,con);
			return;
		}
	}
/* Generate appropriate responces */
	if(httpver){
		consend(con,"HTTP/1.0 200 OK\r\n");
		filetype(path,con);
		sprintf(path,"Server: %s\r\n",myname);
		consend(con,path);
		consend(con,"\r\n");
	}
	buf=(char*)malloc(1024);
	bzero(buf,1024);
/* Send the file in 1K chunks */
	while(!feof(in)){
		i=fread(buf,sizeof(char),1024,in);
		if(write(con->descr,buf,i) < 0){
			disconnect(conhead,con);
			free(buf);
			fclose(in);
			return;
		}
	}
	free(buf);
	disconnect(conhead,con);
	fclose(in); /* Close the file */
}

void filetype(char* path,CON* con){
	int		i;
	char		temp[256];
	FILE*	in;
/* First, let's see if we can locally ID the file.  Otherwise, we'll
 * try the extension.
 */
	in=fopen(path,"r");
/* Read in first 256 bytes */
	if(!in)
		return;
	fread(temp,sizeof(char),256,in);
	fclose(in);
/* Is it a GIF? */
	if((!strncmp(temp,"GIF87a",6))||(!strncmp(temp,"GIF89a",6))){
		consend(con,"content-type: image/gif\r\n");
		return;
	}else if((!strncmp(temp,"\115\115",2))||(!strncmp(temp,"\111\111",2))){
		consend(con,"content-type: image/tiff\r\n");
		return;
	}else if(!strncmp(temp,"\377\330\377\340",2)){
		consend(con,"content-type: image/jpeg\r\n");
		return;
	}
/* Otherwise, we'll check the file extension */
	for(i=0; path[i]!='\0'; i++){
		if(!strncmp(&path[i],".jpg",4)){
			consend(con,"content-type: image/jpeg\r\n");
			return;
		}else if(!strncmp(&path[i],".html",5)){
			consend(con,"content-type: text/html\r\n");
			return;
		}else if(!strncmp(&path[i],".htm",4)){
			consend(con,"content-type: text/html\r\n");
			return;
		}else if(!strncmp(&path[i],".gif",4)){
			consend(con,"content-type: image/gif\r\n");
			return;
		}else if(!strncmp(&path[i],".jpeg",5)){
			consend(con,"content-type: image/jpeg\r\n");
			return;
		}
	}
	consend(con,"content-type: text/plain\r\n");
}

void	outdir(char* path,CON* con,int httpver){
	char*	index;
	char*	tmp;
	int	i;
	FILE*	in;
	DIR*	dir;
	struct	dirent*	ptr;

/* Let's check for index.html and index.htm */
	index=(char*)malloc(strlen(path)+strlen("index.html")+1);
	strcpy(index,path);
	strcat(index,"index.html");
	if(in=fopen(index,"r")){
		if(httpver){
			consend(con,"HTTP/1.0 200 OK\r\ncontent-type: text/html\r\n\r\n");
		}
		free(index);
		index=(char*)malloc(1024);
		while(!feof(in)){ /* Send it just like we did above */
			bzero(index,1024);
			i=fread(index,sizeof(char),1024,in);
			if(write(con->descr,index,i) < 0){
				disconnect(conhead,con);
				free(index);
				fclose(in);
				return;
			}
		}
		fclose(in);
		free(index);
		disconnect(conhead,con);
		return;
	}
	free(index);
	index=(char*)malloc(strlen(path)+strlen("index.htm")+1);
	strcpy(index,path);
	strcat(index,"index.htm");
	if(in=fopen(index,"r")){
		if(httpver){
			consend(con,"HTTP/1.0 200 OK\r\ncontent-type: text/html\r\n\r\n");
		}
		free(index);
		index=(char*)malloc(1024);
		while(!feof(in)){
			bzero(index,1024);
			i=fread(index,sizeof(char),1024,in);
			if(write(con->descr,index,i) < 0){
				disconnect(conhead,con);
				free(index);
				fclose(in);
				return;
			}
		}
		fclose(in);
		free(index);
		disconnect(conhead,con);
		return;
	}
	free(index);
	dir=opendir(path);
	if(!dir){
		if(httpver){
			consend(con,
			"HTTP/1.0 404 Not Found\r\n"
			"Content-type: text/html\r\n"
			"Connection: close\r\n"
			"\r\n");
		}
		consend(con,
"<HTML><TITLE>Error!</TITLE><BODY>\r\n"
"<H2><CENTER>I'm sorry, but I recieved an error -- Error number 404</H2></CENTER>\r\n"
"This means that the file you were searching for wasn't found on this server.\r\n"
"Please make sure that your URL is correct and that the file exists.\r\n");
		disconnect(conhead,con);
		return;
	}
/* Otherwise, let's read the directory and shoot it all out at people */
	ptr=readdir(dir);
	index=(char*)malloc(8192);
	tmp=(char*)malloc(1024);
	sprintf(index,
"<HTML><TITLE>Index of %s</TITLE><BODY>\r\n"
"<LEFT><H2><BOLD>List of files in %s</BOLD></H2></LEFT><HR>\r\n",path,path);
	while(ptr){ /* Until we hit a null dirent, let's output the filenames */
		if(ptr->d_name[0]!='.'){
			sprintf(tmp,"<A HREF=\"%s\">%s</A><BR>\r\n",ptr->d_name,ptr->d_name);
			strcat(index,tmp);
		}
		ptr=readdir(dir);
	}
	strcat(index,"<HR></BODY></HTML>\r\n");
	if(httpver){
		consend(con,
		"HTTP/1.0 200 OK\r\n"
		"Content-type: text/html\r\n"
		"Connection: close\r\n"
		"\r\n");
	}
	consend(con,index);
	free(index);
	free(ptr);
	closedir(dir);
	disconnect(conhead,con);
}
